@@ -1,6 +1,7 @@
+// SPDX-License-Identifier: Apache-2.0
 /*
  * umoci: Umoci Modifies Open Containers' Images
- * Copyright (C) 2016-2019 SUSE LLC.
+ * Copyright (C) 2016-2025 SUSE LLC
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -18,27 +19,31 @@
 package convert
 
 import (
+	"fmt"
+	"maps"
 	"path/filepath"
 	"strings"
 
 	"github.com/apex/log"
-	igen "github.com/openSUSE/umoci/oci/config/generate"
-	"github.com/openSUSE/umoci/third_party/user"
+	"github.com/blang/semver/v4"
+	"github.com/moby/sys/user"
 	ispec "github.com/opencontainers/image-spec/specs-go/v1"
 	rspec "github.com/opencontainers/runtime-spec/specs-go"
-	"github.com/pkg/errors"
+
+	igen "github.com/opencontainers/umoci/oci/config/generate"
 )
 
 // Annotations described by the OCI image-spec document (these represent fields
 // in an image configuration that do not have a native representation in the
 // runtime-spec).
 const (
-	osAnnotation           = "org.opencontainers.image.os"
-	archAnnotation         = "org.opencontainers.image.architecture"
-	authorAnnotation       = "org.opencontainers.image.author"
-	createdAnnotation      = "org.opencontainers.image.created"
-	stopSignalAnnotation   = "org.opencontainers.image.stopSignal"
-	exposedPortsAnnotation = "org.opencontainers.image.exposedPorts"
+	platformOsAnnotation      = "org.opencontainers.image.os"
+	platformArchAnnotation    = "org.opencontainers.image.architecture"
+	platformVariantAnnotation = "org.opencontainers.image.variant"
+	authorAnnotation          = "org.opencontainers.image.author"
+	createdAnnotation         = "org.opencontainers.image.created"
+	stopSignalAnnotation      = "org.opencontainers.image.stopSignal"
+	exposedPortsAnnotation    = "org.opencontainers.image.exposedPorts"
 )
 
 // ToRuntimeSpec converts the given OCI image configuration to a runtime
@@ -59,12 +64,12 @@ func ToRuntimeSpec(rootfs string, image ispec.Image) (rspec.Spec, error) {
 func parseEnv(env string) (string, string, error) {
 	parts := strings.SplitN(env, "=", 2)
 	if len(parts) != 2 {
-		return "", "", errors.Errorf("environment variable must contain '=': %s", env)
+		return "", "", fmt.Errorf("environment variable must contain '=': %s", env)
 	}
 
 	name, value := parts[0], parts[1]
 	if name == "" {
-		return "", "", errors.Errorf("environment variable must have non-empty name: %s", env)
+		return "", "", fmt.Errorf("environment variable must have non-empty name: %s", env)
 	}
 	return name, value, nil
 }
@@ -103,22 +108,39 @@ func allocateNilStruct(spec *rspec.Spec) {
 }
 
 // MutateRuntimeSpec mutates a given runtime configuration with the image
-// configuration provided.
+// configuration provided in accordance with the image specification's
+// conversion mechanism (for more information, see
+// <https://github.com/opencontainers/image-spec/blob/main/conversion.md>).
 func MutateRuntimeSpec(spec *rspec.Spec, rootfs string, image ispec.Image) error {
 	ig, err := igen.NewFromImage(image)
 	if err != nil {
-		return errors.Wrap(err, "creating image generator")
+		return fmt.Errorf("creating image generator: %w", err)
 	}
 
-	if ig.OS() != "linux" {
-		return errors.Errorf("unsupported OS: %s", image.OS)
+	if ig.PlatformOS() != "linux" {
+		return fmt.Errorf("unsupported OS: %s", image.OS)
 	}
 
 	allocateNilStruct(spec)
 
-	// FIXME: We need to figure out if we're modifying an incompatible runtime spec.
-	//spec.Version = rspec.Version
-	spec.Version = "1.0.0"
+	// Default config to our rspec version if none was specified.
+	if spec.Version == "" {
+		spec.Version = curSpecVersion.String()
+	}
+
+	// Make sure that the previous version of the spec is compatible with us.
+	// We cannot operate on specifications that are newer than us (because we
+	// might drop fields that the user finds important).
+	oldVersion, err := semver.Parse(spec.Version)
+	if err != nil {
+		return fmt.Errorf("parsing original runtime-spec config version: %w", err)
+	}
+	if oldVersion.GT(curSpecVersion) {
+		return fmt.Errorf("original runtime-spec config version %s is unsupported: %s > %s", oldVersion, oldVersion, curSpecVersion)
+	}
+	if oldVersion.Major != curSpecVersion.Major {
+		return fmt.Errorf("original runtime-spec config version %s is incompatible with version %s: mismatching major number", oldVersion, curSpecVersion)
+	}
 
 	// Set verbatim fields
 	spec.Process.Terminal = true
@@ -133,7 +155,7 @@ func MutateRuntimeSpec(spec *rspec.Spec, rootfs string, image ispec.Image) error
 	for _, env := range ig.ConfigEnv() {
 		name, value, err := parseEnv(env)
 		if err != nil {
-			return errors.Wrap(err, "parsing image.Config.Env")
+			return fmt.Errorf("parsing image.Config.Env: %w", err)
 		}
 		appendEnv(&spec.Process.Env, name, value)
 	}
@@ -145,15 +167,23 @@ func MutateRuntimeSpec(spec *rspec.Spec, rootfs string, image ispec.Image) error
 		spec.Process.Args = args
 	}
 
-	// Set annotations fields
-	for key, value := range ig.ConfigLabels() {
-		spec.Annotations[key] = value
+	// Set the "annotation fields".
+	setAnnotation := func(name, value string) {
+		if value != "" {
+			spec.Annotations[name] = value
+		} else {
+			delete(spec.Annotations, name)
+		}
 	}
-	spec.Annotations[osAnnotation] = ig.OS()
-	spec.Annotations[archAnnotation] = ig.Architecture()
-	spec.Annotations[authorAnnotation] = ig.Author()
-	spec.Annotations[createdAnnotation] = ig.Created().Format(igen.ISO8601)
-	spec.Annotations[stopSignalAnnotation] = image.Config.StopSignal
+	setAnnotation(platformOsAnnotation, ig.PlatformOS())
+	setAnnotation(platformArchAnnotation, ig.PlatformArchitecture())
+	setAnnotation(platformVariantAnnotation, ig.PlatformVariant())
+	setAnnotation(authorAnnotation, ig.Author())
+	setAnnotation(createdAnnotation, ig.Created().Format(igen.ISO8601))
+	setAnnotation(stopSignalAnnotation, image.Config.StopSignal)
+	setAnnotation(exposedPortsAnnotation, strings.Join(ig.ConfigExposedPorts(), ","))
+	// Config.Labels need to be applied after the auto-applied labels.
+	maps.Copy(spec.Annotations, ig.ConfigLabels())
 
 	// Set parsed fields
 	// Get the *actual* uid and gid of the user. If the image doesn't contain
@@ -169,9 +199,9 @@ func MutateRuntimeSpec(spec *rspec.Spec, rootfs string, image ispec.Image) error
 		// We only log an error if were not given a rootfs, and we set execUser
 		// to the "default" (root:root).
 		if rootfs != "" {
-			return errors.Wrapf(err, "cannot parse user spec: '%s'", ig.ConfigUser())
+			return fmt.Errorf("cannot parse user spec: %q: %w", ig.ConfigUser(), err)
 		}
-		log.Warnf("could not parse user spec '%s' without a rootfs -- defaulting to root:root", ig.ConfigUser())
+		log.Warnf("could not parse user spec %q without a rootfs -- defaulting to root:root", ig.ConfigUser())
 		execUser = new(user.ExecUser)
 	}
 
@@ -187,11 +217,7 @@ func MutateRuntimeSpec(spec *rspec.Spec, rootfs string, image ispec.Image) error
 		appendEnv(&spec.Process.Env, "HOME", execUser.Home)
 	}
 
-	// Set optional fields
-	ports := ig.ConfigExposedPortsArray()
-	spec.Annotations[exposedPortsAnnotation] = strings.Join(ports, ",")
-
-	for vol := range ig.ConfigVolumes() {
+	for _, vol := range ig.ConfigVolumes() {
 		// XXX: This is _fine_ but might cause some issues in the future.
 		spec.Mounts = append(spec.Mounts, rspec.Mount{
 			Destination: vol,
